@@ -8,13 +8,15 @@
 package Gtk2::Ex::Geo::Overlay;
 
 use strict;
-use UNIVERSAL qw(isa);
 use POSIX;
+use Scalar::Util qw(blessed);
 use Carp;
 use Glib qw/TRUE FALSE/;
 use Geo::OGC::Geometry;
 
-our $VERSION = '0.61'; # same as Geo.pm
+use vars qw / $EDIT_SNAP_DISTANCE /;
+
+our $VERSION = '0.62'; # same as Geo.pm
 
 =pod
 
@@ -27,23 +29,20 @@ documentation of Gtk2::Ex::Geo</a> is written in doxygen format.
 
 =cut
 
-my %visual_keys = ($Gtk2::Gdk::Keysyms{plus}=>1,
-		   $Gtk2::Gdk::Keysyms{minus}=>1,
-		   $Gtk2::Gdk::Keysyms{Right}=>1,
-		   $Gtk2::Gdk::Keysyms{Left}=>1,
-		   $Gtk2::Gdk::Keysyms{Up}=>1,
-		   $Gtk2::Gdk::Keysyms{Down}=>1);
-
+$EDIT_SNAP_DISTANCE = 5;
 
 use Glib::Object::Subclass
     Gtk2::ScrolledWindow::,
     signals => {
-	update_layers => {}, # sent just before the layers are rendered
-	new_selection => {}, # sent after attribute {selection} has changed
-	zoomed_in => {},     # deprecated
-	extent_changed => {},# deprecated
-	motion_notify => {}, # the mouse has a new location on the map
-	map_updated => {},   # deprecated
+	update_layers => {},  # sent just before the layers are rendered
+	new_selection => {},  # sent after the user has changed the selection
+	drawing_changed => {},# sent after the user has changed the drawing
+	zoomed_in => {},      # deprecated
+	extent_changed => {}, # deprecated
+	motion_notify => {},  # the mouse has a new location on the map
+	map_updated => {},    # deprecated
+	pixmap_ready => {},   # sent just after pixmap is ready, but selection and drawing 
+	                      # haven't been rendered, connect to this for annotations
     },
     properties => 
     [
@@ -87,25 +86,28 @@ sub INIT_INSTANCE {
     $self->{rubberband_geometry} = '';
     $self->{rubberband_mode} = '';
         
-    # why do I need to set these?
+    # why do I need to set these two?
     $self->{zoom_factor} = 1.2;
     $self->{step} = 8;
 
     $self->{offset} = [0, 0];
-    @{$self->{bg_color}} = (0, 0, 0);
+    $self->{bg_color} = [255, 255, 255, 255];
+    $self->{selection_color} = [255*257, 178*257, 0];
+    $self->{selection_style} = 'GDK_LINE_SOLID';
+    $self->{drawing_color} = [0, 65535, 0];
+    $self->{show_selection} = 1;
+
+    $self->{layers} = [];
 }
 
 ## @method
 # @brief Attempt to delete all widgets within this widget.
 sub close {
     my $self = shift;
-    delete $self->{image};
-    delete $self->{event_box};
-    delete $self->{pixmap};
-    delete $self->{pixbuf};
-    delete $self->{old_hadj};
-    delete $self->{old_vadj};
-    delete $self->{rubberband_gc};
+    while (my($key, $widget) = each %$self) {
+	$widget->destroy if blessed($widget) and $widget->isa("Gtk2::Widget");
+	delete $self->{$key};
+    }
 }
 
 ## @ignore
@@ -119,7 +121,9 @@ sub size_allocate {
     return 0;
 }
 
-## @ignore
+## @method my_inits
+# @brief call after new
+# @todo merge into new
 sub my_inits {
     my($self, %params) = @_;
     $self->{inited} = 1;
@@ -130,6 +134,10 @@ sub my_inits {
     $self->add_with_viewport($self->{event_box});
 
     for (keys %params) {
+	if ($_ eq 'bg_color' or $_ eq 'offset') {
+	    @{$self->{$_}} = @{$params{$_}};
+	    next;
+	}
 	$self->{$_} = $params{$_};
     }
 }
@@ -140,8 +148,7 @@ sub my_inits {
 sub add_layer {
     my($self, $layer, $do_not_zoom_to) = @_;
 
-    croak "blocked an attempt to put a non Gtk2::Ex::Geo::Layer into overlay" unless 
-	isa($layer, 'Gtk2::Ex::Geo::Layer');
+    return unless blessed($layer) and $layer->isa('Gtk2::Ex::Geo::Layer');
 
     push @{$self->{layers}}, $layer;
 
@@ -241,7 +248,10 @@ sub zoom_to {
     if (@_ == 1) {
 	my $layer = shift;
 	return unless $self->{layers} and @{$self->{layers}};
-	@bounds = $layer->world(of_GDAL=>1);
+	eval {
+	    @bounds = $layer->world();
+	};
+	@bounds = (0,0,1,1) unless @bounds;
 	$self->{offset} = [0, 0];
     } elsif (@_ == 5) {
 	my($minX, $maxY, $pixel_size, @offset) = @_;
@@ -281,7 +291,7 @@ sub zoom_to {
     }
 }
 
-## @method get_world
+## @method get_world()
 # @brief Get the total area of the canvas.
 # @return (min_x, min_y, max_x, max_y)
 sub get_world {
@@ -289,7 +299,7 @@ sub get_world {
     return ($self->{minX}, $self->{minY}, $self->{maxX}, $self->{maxY});
 }
 
-## @method get_viewport
+## @method get_viewport()
 # @brief Get the visible area of the canvas.
 # @return (min_x, min_y, max_x, max_y)
 sub get_viewport {
@@ -301,7 +311,7 @@ sub get_viewport {
 	     $minX+$self->{viewport_size}->[0]*$self->{pixel_size}, $maxY );
 }
 
-## @method get_viewport_of_selection
+## @method get_viewport_of_selection()
 # @brief Get the visible area of the canvas.
 # @return (min_x, min_y, max_x, max_y)
 sub get_viewport_of_selection {
@@ -313,21 +323,25 @@ sub get_viewport_of_selection {
     return ($ll->X, $ll->Y, $ur->X, $ur->Y);
 }
 
-## @method size
+## @method size()
 # @brief The size of the viewport in pixels (height, width)
 sub size {
     my $self = shift;
     return ($self->{viewport_size}->[1], $self->{viewport_size}->[0]);
 }
 
-## @method zoom_to_all
+## @method zoom_to_all()
 # @brief Sets the world as the bounding box for all layers
 sub zoom_to_all {
     my($self) = @_;
     return unless $self->{layers} and @{$self->{layers}};
     my @size;
     for my $layer (@{$self->{layers}}) {
-	my @s = $layer->world(of_GDAL=>1);
+	my @s;
+	eval {
+	    @s = $layer->world();
+	};
+	next unless @s;
 	if (@size) {
 	    $size[0] = min($size[0], $s[0]);
 	    $size[1] = min($size[1], $s[1]);
@@ -340,18 +354,6 @@ sub zoom_to_all {
     $self->zoom_to(@size) if @size;
 }
 
-## @method set_draw_on($draw_on, $user_param)
-# @brief Sets the method that is called to annotate the pixmap
-# the $draw_on sub is called like this:
-# @code
-# $draw_on->($user_param, $pixmap);
-# @endcode
-sub set_draw_on {
-    my($self, $draw_on, $user_param) = @_;
-    $self->{draw_on} = $draw_on;
-    $self->{draw_on_user_param} = $user_param;
-}
-
 ## @ignore
 sub value_changed {
     my(undef, $self) = @_;
@@ -362,7 +364,7 @@ sub value_changed {
     return 1;
 }
 
-## @method get_focus
+## @method get_focus()
 # @deprecated use get_viewport_of_selection or get_viewport
 # @returns the visible area or the selection, if one exists, as ($minx, $miny, $maxx, $maxy).
 sub get_focus {
@@ -381,6 +383,23 @@ sub get_focus {
 }
 
 {
+    package Gtk2::Ex::Geo::PseudoOverlay;
+    sub round {
+	return int($_[0] + .5 * ($_[0] <=> 0));
+    }
+    sub new {
+	my($class, $minX, $maxY, $pixel_size) = @_;	
+	my $self = { minX => $minX,
+		     maxY => $maxY,
+		     pixel_size => $pixel_size
+	};
+	bless($self, $class); 
+    }
+    sub point2pixmap_pixel {
+	my($self, @p) = @_;
+	return (round(($p[0] - $self->{minX})/$self->{pixel_size} - 0.5),
+		round(($self->{maxY} - $p[1])/$self->{pixel_size} - 0.5));
+    }
     package Gtk2::Ex::Geo::Canvas;
     our @ISA = qw(Gtk2::Gdk::Pixbuf);
  
@@ -388,9 +407,11 @@ sub get_focus {
 	my($class, $layers, 
 	   $minX, $maxY, $pixel_size, $w_offset, $h_offset,
 	   $width, $height,
-	   $bg_r, $bg_g, $bg_b, $overlay) = @_;
+	   $bg_r, $bg_g, $bg_b, $bg_a, $overlay) = @_;
 	
 	return unless defined $minX;
+
+	$overlay = Gtk2::Ex::Geo::PseudoOverlay->new($minX, $maxY, $pixel_size) unless $overlay;
 	
 	my @viewport = ($minX+$pixel_size*$w_offset, 0, 0, $maxY-$pixel_size*$h_offset);
 	$viewport[2] = $viewport[0]+$pixel_size*$width;
@@ -399,7 +420,7 @@ sub get_focus {
 	my $pb = &Gtk2::Ex::Geo::gtk2_ex_geo_pixbuf_create($width, $height,
 							   $viewport[0], $viewport[3],
 							   $pixel_size, 
-							   $bg_r, $bg_g, $bg_b);
+							   $bg_r, $bg_g, $bg_b, $bg_a);
 	
 	my $surface = &Gtk2::Ex::Geo::gtk2_ex_geo_pixbuf_get_cairo_surface($pb);
 	my $cr = Cairo::Context->create($surface);
@@ -417,6 +438,8 @@ sub get_focus {
     }
 }
 
+package Gtk2::Ex::Geo::Overlay;
+
 ## @method render(%params)
 # @brief Render the layers on the canvas.
 # Each layer's render method is called:
@@ -433,13 +456,14 @@ sub render {
     my $self = shift;
     my %opt = @_;
 
-    return unless $self->{viewport_size}->[0];
+    my $size = $self->{viewport_size};
+    return unless $size->[0];
 
     $self->signal_emit('update-layers');
 
     my @tmp = ($self->{minX}, $self->{maxY}, $self->{pixel_size}, @{$self->{offset}});
     $self->{pixbuf} = Gtk2::Ex::Geo::Canvas->new
-	($self->{layers}, @tmp, @{$self->{viewport_size}}, @{$self->{bg_color}}, $self);
+	($self->{layers}, @tmp, @{$size}, @{$self->{bg_color}}[0..3], $self);
 
     return unless $self->{pixbuf};
 
@@ -457,13 +481,13 @@ sub render {
 
     $self->{old_hadj} = $self->get_hscrollbar->get_adjustment; # prevents a warning
     $self->get_hscrollbar->set_adjustment
-	(Gtk2::Adjustment->new($self->{offset}[0], 0, $self->{canvas_size}[0], $self->{viewport_size}[0]/20,
-			       $self->{viewport_size}[0], $self->{viewport_size}[0]));
+	(Gtk2::Adjustment->new($self->{offset}[0], 0, $self->{canvas_size}[0], $size->[0]/20,
+			       $size->[0], $size->[0]));
 
     $self->{old_vadj} = $self->get_vscrollbar->get_adjustment; # prevents a warning
     $self->get_vscrollbar->set_adjustment
-	(Gtk2::Adjustment->new($self->{offset}[1], 0, $self->{canvas_size}[1], $self->{viewport_size}[1]/20,
-			       $self->{viewport_size}[1], $self->{viewport_size}[1]));
+	(Gtk2::Adjustment->new($self->{offset}[1], 0, $self->{canvas_size}[1], $size->[1]/20,
+			       $size->[1], $size->[1]));
 
     $self->signal_emit ('map-updated');
 
@@ -477,61 +501,77 @@ sub render {
 # @param gc A gdk graphics context (Gtk2::Gdk::GC object)
 # @param geom A Geo::OGC::Geometry object.
 sub render_geometry {
-    my($self, $gc, $geom) = @_;
-    if (isa($geom, 'Geo::OGC::GeometryCollection')) 
+    my($self, $gc, $geom, %param) = @_;
+    if ($geom->isa('Geo::OGC::GeometryCollection')) 
     {
 	for my $g ($geom->NumGeometries) {
-	    $self->render_geometry($gc, $g);
+	    $self->render_geometry($gc, $g, %param);
 	}
 	return;
     } 
-    elsif (isa($geom, 'Geo::OGC::Point')) 
+    elsif ($geom->isa('Geo::OGC::Point')) 
     {
 	my @p = $self->point2pixmap_pixel($geom->X, $geom->Y);
 	$self->{pixmap}->draw_line($gc, $p[0]-4, $p[1], $p[0]+4, $p[1]);
 	$self->{pixmap}->draw_line($gc, $p[0], $p[1]-4, $p[0], $p[1]+4);
     } 
-    elsif (isa($geom, 'Geo::OGC::LineString')) 
+    elsif ($geom->isa('Geo::OGC::LineString')) 
     {
 	my @points;
 	for my $p ($geom->NumPoints) {
 	    push @points, $self->point2pixmap_pixel($p->X, $p->Y);
 	}
 	$self->{pixmap}->draw_lines($gc, @points);
+	if ($param{enhance_vertices}) {
+	    my $pm = $self->{pixmap};
+	    for (my $i = 0; $i < $#points; $i+=2) {
+		my $x = $points[$i];
+		my $y = $points[$i+1];
+		$pm->draw_line($gc, $x-4, $y, $x+4, $y);
+		$pm->draw_line($gc, $x, $y-4, $x, $y+4);
+	    }
+	}
     }
-    elsif (isa($geom, 'Geo::OGC::Polygon')) 
+    elsif ($geom->isa('Geo::OGC::Polygon')) 
     {
-	$self->render_geometry($gc, $geom->ExteriorRing);
+	$self->render_geometry($gc, $geom->ExteriorRing, %param);
+	for my $i (0..$geom->NumInteriorRing-1) {
+	    $self->render_geometry($gc, $geom->InteriorRingN($i), %param);
+	}
     }
 }
 
-sub reset_pixmap {
-    my($self) = @_;
-    $self->{pixmap} = $self->{pixbuf}->render_pixmap_and_mask(0);
-}
-
-sub reset_image {
-    my($self) = @_;
+## @method update_image($annotations, $user_param)
+# @param annotations A subroutine for user annotations. Called like
+# this: $annotations->($overlay, $pixmap, $gc, $user_param).
+# @param user_param User parameter for the annotations.
+# @brief Updates the image on the screen to show the changes in pixmap.
+sub update_image {
+    my($self, $annotations, $user_param) = @_;
+    return unless $self->{pixbuf};
     $self->{image}->set_from_pixbuf(undef);
-    if ($self->{selection}) {
-	my $gc = Gtk2::Gdk::GC->new($self->{pixmap});
-	$gc->set_rgb_fg_color(Gtk2::Gdk::Color->new(65535, 65535, 0));
-	my $style = 'GDK_LINE_SOLID'; # unless in collection each geom can have their own style
+    $self->{pixmap} = $self->{pixbuf}->render_pixmap_and_mask(0);
+    my $gc = Gtk2::Gdk::GC->new($self->{pixmap});
+    $self->{pixmap}->draw_line($gc, 0, 0, 0, 0); # strange bug, the first line is not drawn
+    $self->signal_emit('pixmap_ready');
+    my $first = 1;
+    if ($self->{drawing}) {
+	$gc->set_rgb_fg_color(Gtk2::Gdk::Color->new(@{$self->{drawing_color}}));
+	my $style = 'GDK_LINE_SOLID';
+	$gc->set_line_attributes(2, $style, 'GDK_CAP_NOT_LAST', 'GDK_JOIN_MITER');
+	$self->render_geometry($gc, $self->{drawing}, enhance_vertices => 1);
+    }
+    if ($self->{selection} and $self->{show_selection}) {
+	$gc->set_rgb_fg_color(Gtk2::Gdk::Color->new(@{$self->{selection_color}}));
+	my $style = $self->{selection_style};
+	$style = 'GDK_LINE_SOLID';
 	$gc->set_line_attributes(2, $style, 'GDK_CAP_NOT_LAST', 'GDK_JOIN_MITER');
 	$self->render_geometry($gc, $self->{selection});
     }
+    $annotations->($self, $self->{pixmap}, $gc, $user_param) if $annotations;
     $self->{image}->set_from_pixmap($self->{pixmap}, undef);
 }
 
-## @method update_image
-# @brief Updates the image on the screen to show the changes in pixmap.
-sub update_image {
-    my($self) = @_;
-    $self->reset_pixmap;
-    $self->{draw_on}->($self->{draw_on_user_param}, $self->{pixmap}) if $self->{draw_on};
-    $self->reset_image;
-}
-
 ## @method zoom($w_offset, $h_offset, $pixel_size)
 # @brief Select a part of the world into the visible area.
 sub zoom {
@@ -631,7 +671,7 @@ sub pan {
 # @brief Handling of key press events.
 #
 # Tied to key_press_event and key_release_event. Ties "+" to zoom_in,
-# "-" to zoom_out,and arrow keysto pan. Also ties "Esc" to finishing
+# "-" to zoom_out,and arrow keysto pan. Also ties "Enter" to finishing
 # making a selection. Records press and release of "Ctrl" to object
 # attribute "_control_down".
 sub key_press_event {
@@ -639,14 +679,12 @@ sub key_press_event {
 
     return 0 unless $self->{layers} and @{$self->{layers}};
     
-    # if this were an event box handler like button press
-#    my(undef, $event, $self) = @_;
-
     my $key = $event->keyval;
-    if ($key == $Gtk2::Gdk::Keysyms{plus}) {
+    #print STDERR "key=$key\n";
+    if ($key == $Gtk2::Gdk::Keysyms{plus}) {	    
 	$self->zoom_in($event); # , $self->event_pixel2point());
     } elsif ($key == $Gtk2::Gdk::Keysyms{minus}) {
-	$self->zoom_out($event); # , $self->event_pixel2point());
+	$self->zoom_out($event); # , $self->event_pixel2point());	
     } elsif ($key == $Gtk2::Gdk::Keysyms{Right}) {
 	$self->pan($self->{viewport_size}->[0]/$self->{step}, 0, $event);
     } elsif ($key == $Gtk2::Gdk::Keysyms{Left}) {
@@ -656,7 +694,9 @@ sub key_press_event {
     } elsif ($key == $Gtk2::Gdk::Keysyms{Down}) {
 	$self->pan(0, $self->{viewport_size}->[1]/$self->{step}, $event);
     } elsif ($key == $Gtk2::Gdk::Keysyms{Escape}) {
-	if ($self->{rubberband_mode} eq 'select' and $self->{path}) {
+	$self->delete_rubberband;
+    } elsif ($key == $Gtk2::Gdk::Keysyms{Return}) {
+	if ($self->draw_mode and $self->{path}) {
 	    if ($self->{rubberband_geometry} eq 'polygon') {
 		if (@{$self->{path}} > 2) {
 		    my $geom = new Geo::OGC::Polygon;
@@ -680,13 +720,54 @@ sub key_press_event {
 	    }
 	}
 	$self->delete_rubberband;
-    } elsif ($key == $Gtk2::Gdk::Keysyms{Control_L} or $key == $Gtk2::Gdk::Keysyms{Control_R}) {
+
+    } elsif ($key == $Gtk2::Gdk::Keysyms{Insert}) {
+
+	if ($self->{rubberband_mode} eq 'edit' and $self->{drawing}) {
+	    
+	    # find the closest point in drawing
+	    my @p = $self->event_pixel2point(@{$self->{event_coordinates}});
+	    my @r = $self->{drawing}->ClosestPoint(@p);
+	    my $d = pop @r;
+	    if (@r and $d/$self->{pixel_size} < $EDIT_SNAP_DISTANCE) {
+		$self->{drawing}->AddVertex(@r);
+		$self->signal_emit('drawing-changed');
+		$self->update_image;
+	    }
+	}
+	
+    } elsif ($key == $Gtk2::Gdk::Keysyms{Delete}) {
+
+	if ($self->{rubberband_mode} eq 'edit' and $self->{drawing}) {
+	    
+	    # find the closest vertex in drawing
+	    my @p = $self->event_pixel2point(@{$self->{event_coordinates}});
+	    my @r = $self->{drawing}->ClosestVertex(@p);
+	    my $d = pop @r;
+	    if (@r and $d/$self->{pixel_size} < $EDIT_SNAP_DISTANCE) {
+		$self->{drawing}->DeleteVertex(@r);
+		$self->signal_emit('drawing-changed');
+		$self->update_image;
+	    }	    
+	}
+
+    } elsif (($key == $Gtk2::Gdk::Keysyms{Control_L} or $key == $Gtk2::Gdk::Keysyms{Control_R}) and
+	     $self->draw_mode) {
 	$self->{_control_down} = 1;
-	return 0;
+    } elsif (($key == $Gtk2::Gdk::Keysyms{Shift_L} or $key == $Gtk2::Gdk::Keysyms{Shift_R}) and
+	     $self->draw_mode and
+	     ($self->{drawing} and $self->{drawing}->LastPolygon)) {
+	$self->{_shift_down} = 1;
     }
     return 0;
 }
 
+## @ignore
+sub draw_mode {
+    my($self) = @_;
+    return ($self->{rubberband_mode} eq 'select' or $self->{rubberband_mode} eq 'draw');
+}
+
 ## @method key_release_event($event)
 # @brief Handling of key release events.
 #
@@ -694,26 +775,39 @@ sub key_press_event {
 sub key_release_event {
     my($self, $event) = @_;
     my $key = $event->keyval;
-    if ($key == $Gtk2::Gdk::Keysyms{Control_L} or 
-	$key == $Gtk2::Gdk::Keysyms{Control_R}) {
+    if ($key == $Gtk2::Gdk::Keysyms{Control_L} or $key == $Gtk2::Gdk::Keysyms{Control_R}) {
 	$self->{_control_down} = 0;
+    } elsif ($key == $Gtk2::Gdk::Keysyms{Shift_L} or $key == $Gtk2::Gdk::Keysyms{Shift_R}) {
+	$self->{_shift_down} = 0;
     }
 }
 
 sub add_to_selection {
     my($self, $geom) = @_;
+    my $store = ($self->{rubberband_mode} eq 'select') ? 'selection' : 'drawing';
     if ($self->{_control_down}) {
-	if (!$self->{selection} or
-	    !isa($self->{selection}, 'Geo::OGC::GeometryCollection')) {
+	
+	# create first a multi something, then fall back to collection if need be
+
+	unless ($self->{$store} and $self->{$store}->isa('Geo::OGC::GeometryCollection')) {
 	    my $coll = Geo::OGC::GeometryCollection->new;
-	    $coll->AddGeometry($self->{selection}) if $self->{selection};
-	    $self->{selection} = $coll;
+	    $coll->AddGeometry($self->{$store}) if $self->{$store};
+	    $self->{$store} = $coll;
+	}
+	$self->{$store}->AddGeometry($geom) if $geom;
+    } elsif ($self->{_shift_down}) {
+	my $polygon = $self->{$store}->LastPolygon;
+	if ($polygon and $geom->isa('Geo::OGC::Polygon')) {
+	    $geom = $geom->ExteriorRing;
+	    # exterior is ccw, interior is cw
+	    $geom->Rotate;
+	    $polygon->AddInteriorRing($geom);
 	}
-	$self->{selection}->AddGeometry($geom) if $geom;
     } else {
-	$self->{selection} = $geom;
+	$self->{$store} = $geom;
     }
-    $self->signal_emit('new_selection');
+    $self->signal_emit('new_selection') if $self->{rubberband_mode} eq 'select';
+    $self->signal_emit('drawing-changed') if $self->{rubberband_mode} eq 'draw';
 }
 
 ## @method button_press_event()
@@ -728,16 +822,16 @@ sub button_press_event {
 
     if ($event->button == 3 and $self->{menu}) {
 
-	$self->delete_rubberband;
+	$self->delete_rubberband if $self->{path};
 	my $menu = Gtk2::Menu->new;
-	for (sort {$self->{menu}{$a}{nr} <=> $self->{menu}{$b}{nr}} keys %{$self->{menu}}) {
-	    my $name = $self->{menu_item_setup}->($_, $self);
+	for (my $i =0; $i < @{$self->{menu}}; $i+=2) {
+	    my $name = $self->{menu_item_setup}->($self->{menu}->[$i], $self);
 	    my $item;
-	    unless ($self->{menu}{$_}{sub}) {
+	    unless ($self->{menu}->[$i+1]) {
 		$item = Gtk2::SeparatorMenuItem->new();
 	    } else {
 		$item = Gtk2::MenuItem->new($name);
-		$item->signal_connect(activate => $self->{menu}{$_}{sub}, $self);
+		$item->signal_connect(activate => $self->{menu}->[$i+1], $self);
 	    }
 	    $item->show;
 	    $menu->append ($item);
@@ -749,13 +843,29 @@ sub button_press_event {
 
 	push @{$self->{path}}, [$event->x, $event->y];
 
-	$self->{rubberband_gc} = Gtk2::Gdk::GC->new ($self->{pixmap});
-	$self->{rubberband_gc}->copy($self->style->fg_gc($self->state));
-	$self->{rubberband_gc}->set_function('invert');
+	unless ($self->{rubberband_mode} eq 'edit') {
+	    $self->{rubberband_gc} = Gtk2::Gdk::GC->new($self->{pixmap});
+	    $self->{rubberband_gc}->copy($self->style->fg_gc($self->state));
+	    $self->{rubberband_gc}->set_function('invert');
+	}
 
-	if (($self->{rubberband_mode} eq 'select' and !$self->{_control_down})) {
-	    delete $self->{selection};
-	    $self->signal_emit('new_selection');
+	if ($self->{rubberband_mode} eq 'edit' and $self->{drawing}) {
+	    # find the closest point in drawing
+	    my @p = $self->event_pixel2point($event->x, $event->y);
+	    my @q = $self->{drawing}->ClosestVertex(@p);
+	    my $d = pop @q;
+	    $self->{drawing_edit} = \@q if $d/$self->{pixel_size} < $EDIT_SNAP_DISTANCE;
+	} elsif ($self->draw_mode and !($self->{_control_down} or $self->{_shift_down}) and
+		 !($self->{rubberband_geometry} eq 'polygon' or $self->{rubberband_geometry} eq 'path')
+	    )
+	{
+	    if ($self->{rubberband_mode} eq 'select') {
+		delete $self->{selection};
+		#$self->signal_emit('new_selection');
+	    } elsif ($self->{rubberband_mode} eq 'draw') {
+		delete $self->{drawing};
+		#$self->signal_emit('drawing-changed');
+	    }
 	}
 
 	$handled = 1;
@@ -765,6 +875,167 @@ sub button_press_event {
     return $handled;
 }
 
+## @method motion_notify()
+# @brief Updates the rubberband if rubberbanding.
+# @todo Use more visible rubberband, there's no need to use XOR.
+sub motion_notify {
+    my(undef, $event, $self) = @_;
+
+    return 0 unless $self->{layers} and @{$self->{layers}};
+
+    @{$self->{event_coordinates}} = ($event->x, $event->y);
+
+    unless ($self->{path}) {
+	$self->signal_emit('motion-notify');
+	return 0; # not handled
+    }
+
+    my $pm = $self->{pixmap};
+    my $rgc = $self->{rubberband_gc};
+    my @begin = @{$self->{path}[0]};
+    my @end = @{$self->{event_coordinates}};
+    my $w = $end[0] - $begin[0];
+    my $h = $end[1] - $begin[1];
+    my @rb = @{$self->{rubberband}} if $self->{rubberband};
+    
+    if ($self->{drawing_edit}) {
+	
+	$pm = $self->{pixmap} = $self->{pixbuf}->render_pixmap_and_mask(0);
+	my @wend = $self->event_pixel2point(@end);
+	my @p = @{$self->{drawing_edit}};
+	my @q = $self->{drawing}->VertexAt(@p);
+	for my $q (@q) {
+	    $q->{X} = $wend[0];
+	    $q->{Y} = $wend[1];
+	}
+	my $gc = Gtk2::Gdk::GC->new($self->{pixmap});
+	$gc->set_rgb_fg_color(Gtk2::Gdk::Color->new(@{$self->{drawing_color}}));
+	my $style = 'GDK_LINE_SOLID';
+	$gc->set_line_attributes(2, $style, 'GDK_CAP_NOT_LAST', 'GDK_JOIN_MITER');
+	$self->render_geometry($gc, $self->{drawing}, enhance_vertices => 1);
+
+    } else {    
+
+	for ($self->{rubberband_mode}.' '.$self->{rubberband_geometry}) {
+	    /edit/ && do {
+		last;
+	    };
+	    /pan/ && do {
+		my $gc = Gtk2::Gdk::GC->new($pm);
+		$pm->draw_rectangle($gc, 1, 0, 0, @{$self->{viewport_size}});
+		$pm->draw_pixbuf($gc, $self->{pixbuf}, 0, 0, $w, $h, -1, -1, 'GDK_RGB_DITHER_NONE', 0, 0);
+		last;
+	    };
+	    /line/ && do {
+		$pm->draw_line($rgc, @rb) if @rb;
+		@rb = (@begin, @end);
+		$pm->draw_line($rgc, @rb);
+	    };
+	    /path/ && do {
+		my @p = @{$self->{path}}; 
+		for my $p (0..$#p-1) {
+		    $pm->draw_line($rgc, @{$p[$p]}, @{$p[$p+1]});
+		}
+		$pm->draw_line($rgc, @rb) if @rb;
+		@rb = (@{$p[$#p]}, @end);
+		for my $p (0..$#p-1) {
+		    $pm->draw_line($rgc, @{$p[$p]}, @{$p[$p+1]});
+		}
+		$pm->draw_line($rgc, @rb);
+	    };
+	    /rect/ && do {
+		$pm->draw_rectangle($rgc, FALSE, @rb) if @rb;
+		@rb = (min($begin[0], $end[0]), min($begin[1], $end[1]), abs($w), abs($h));
+		$pm->draw_rectangle($rgc, FALSE, @rb);
+	    };
+	    /ellipse/ && do {
+		$pm->draw_arc($rgc, FALSE, @rb, 0, 64*360) if @rb;
+		my $a = abs(floor($w * sqrt(2)));
+		my $b = abs(floor($h * sqrt(2)));
+		@rb = ($begin[0] - $a, $begin[1] - $b, 2*$a, 2*$b);
+		$pm->draw_arc($rgc, FALSE, @rb, 0, 64*360);
+	    };
+	    /polygon/ && do {
+		my @p = @{$self->{path}};
+		if (@p == 1) {
+		    $pm->draw_line($rgc, @rb) if @rb;
+		    @rb = (@begin, @end);
+		    $pm->draw_line($rgc, @rb);
+		} else {
+		    $pm->draw_line($rgc, @rb) if @rb and @rb == 4 and @p == 2;
+		    my @points;
+		    for my $p (@p) {
+			push @points, @$p;
+		    }
+		    push @points, @rb if @rb;
+		    $pm->draw_polygon($rgc, 1, @points);
+		    @rb = @end;
+		    pop @points;
+		    pop @points;
+		    push @points, @rb;
+		    $pm->draw_polygon($rgc, 1, @points);
+		}
+	    }
+	}
+    }
+    
+    @{$self->{rubberband}} = @rb;
+
+    $self->{image}->set_from_pixbuf(undef);
+    $self->{image}->set_from_pixmap($pm, undef);
+    
+    $self->signal_emit('motion-notify');
+    return 1; # handled
+}
+
+sub rubberband_mode {
+    my($self) = @_;
+    my $mode = $self->{rubberband_mode};
+    $mode .= ", add to collection" if $self->{_control_down};
+    $mode .= ", add a hole" if $self->{_shift_down};
+    return $mode;
+}
+
+## @method @rubberband_value()
+# @brief Computes a value relevant to current rubberband (length or area) in world coordinates.
+# @return ($dimension, $value) $dimension is either 1 or 2
+sub rubberband_value {
+    my($self) = @_;
+
+    if ($self->{path}) {
+
+	my @p0 = $self->event_pixel2point(@{$self->{path}[0]}) if $self->{path}[0];
+	my @p1 = $self->event_pixel2point(@{$self->{event_coordinates}});
+
+	for ($self->{rubberband_geometry}) {
+	    (/line/ || /path/) && do {
+		my $ogc = new Geo::OGC::LinearRing(points => $self->{path});
+		$ogc->AddPoint(Geo::OGC::Point->new(@{$self->{event_coordinates}}));
+		$ogc->ApplyTransformation( sub { return $self->event_pixel2point(@_); } );
+		return (1, $ogc->Length);
+	    };
+	    /rect/ && do {
+		return (2, abs(($p1[0]-$p0[0])*($p1[1]-$p0[1])));
+	    };
+	    /ellipse/ && do {
+		my $a = ($p1[0]-$p0[0]) * sqrt(2);
+		my $b = ($p1[1]-$p0[1]) * sqrt(2);
+		return (2, abs(3.14159266*$a*$b));
+	    };
+	    /polygon/ && do {
+		my $ogc = new Geo::OGC::LinearRing(points => $self->{path});
+		$ogc->AddPoint(Geo::OGC::Point->new(@{$self->{event_coordinates}}));
+		$ogc->ApplyTransformation( sub { return $self->event_pixel2point(@_); } );
+		$ogc->Close;
+		return (2, undef) unless $ogc->IsSimple;
+		return (2, abs($ogc->Area));
+	    };
+	}
+    }
+}
+
+## @method delete_rubberband()
+# @brief Escapes from rubberbanding.
 sub delete_rubberband {
     my $self = shift;
     delete $self->{path};
@@ -814,7 +1085,7 @@ sub button_release_event {
 		    $self->zoom($w_offset, $h_offset, $pixel_size, 1);
 		}
 	    };
-	    /select/ && do {
+	    (/select/ or /draw/) && do {
 		if ($self->{rubberband_geometry} eq 'line') {
 		    my $geom;
 		    if ($click) {
@@ -826,7 +1097,7 @@ sub button_release_event {
 		    }
 		    $self->add_to_selection($geom);
 		    $self->delete_rubberband;
-		} elsif ($self->{rubberband_geometry} eq 'rect') {
+		} elsif ($self->{rubberband_geometry} =~ /rect/) {
 		    my $geom;
 		    if ($click) {
 			$geom = Geo::OGC::Point->new($wbegin[0], $wbegin[1]);
@@ -851,10 +1122,19 @@ sub button_release_event {
 		    delete $self->{rubberband};
 		}
 	    };
-	    /measure/ && do {
-		if ($self->{rubberband_geometry} eq 'line') {
+	    (/measure/ or /edit/) && do {		
+		if ($self->{drawing_edit}) {
+		    my @p = @{$self->{drawing_edit}};
+		    my @q = $self->{drawing}->VertexAt(@p);
+		    for my $q (@q) {
+			$q->{X} = $wend[0];
+			$q->{Y} = $wend[1];
+		    }
+		    delete $self->{drawing_edit};
+		    $self->delete_rubberband;
+		} elsif ($self->{rubberband_geometry} eq 'line') {
 		    $self->delete_rubberband;		    
-		} elsif ($self->{rubberband_geometry} eq 'rect') {
+		} elsif ($self->{rubberband_geometry} eq 'rectangle') {
 		    $self->delete_rubberband;
 		} elsif ($self->{rubberband_geometry} eq 'ellipse') {
 		    $self->delete_rubberband;
@@ -867,130 +1147,9 @@ sub button_release_event {
     return $handled;
 }
 
-## @method motion_notify
-# @brief Updates the rubberband if rubberbanding.
-# @todo Use more visible rubberband, there's no need to use XOR.
-sub motion_notify {
-    my(undef, $event, $self) = @_;
-
-    return 0 unless $self->{layers} and @{$self->{layers}};
-
-    @{$self->{event_coordinates}} = ($event->x, $event->y);
-
-    my $handled = 0;
-    if ($self->{path}) {
-
-	my $pm = $self->{pixmap};
-	my $rgc = $self->{rubberband_gc};
-	my @begin = @{$self->{path}[0]};
-	my @end = @{$self->{event_coordinates}};
-	my $w = $end[0] - $begin[0];
-	my $h = $end[1] - $begin[1];
-	my @rb = @{$self->{rubberband}} if $self->{rubberband};
-
-	for ($self->{rubberband_mode}.' '.$self->{rubberband_geometry}) {
-	    /pan/ && do {
-		my $gc = new Gtk2::Gdk::GC $pm;
-		$pm->draw_rectangle($gc, 1, 0, 0, @{$self->{viewport_size}});
-		$pm->draw_pixbuf($gc, $self->{pixbuf}, 0, 0, $w, $h, -1, -1, 'GDK_RGB_DITHER_NONE', 0, 0);
-		last;
-	    };
-	    /line/ && do {
-		$pm->draw_line($rgc, @rb) if @rb;
-		@rb = (@begin, @end);
-		$pm->draw_line($rgc, @rb);
-	    };
-	    /path/ && do {
-		my @p = @{$self->{path}}; 
-		for my $p (0..$#p-1) {
-		    $pm->draw_line($rgc, @{$p[$p]}, @{$p[$p+1]});
-		}
-		$pm->draw_line($rgc, @rb) if @rb;
-		@rb = (@{$p[$#p]}, @end);
-		for my $p (0..$#p-1) {
-		    $pm->draw_line($rgc, @{$p[$p]}, @{$p[$p+1]});
-		}
-		$pm->draw_line($rgc, @rb);
-	    };
-	    /rect/ && do {
-		$pm->draw_rectangle($rgc, FALSE, @rb) if @rb;
-		@rb = (min($begin[0], $end[0]), min($begin[1], $end[1]), abs($w), abs($h));
-		$pm->draw_rectangle($rgc, FALSE, @rb);
-	    };
-	    /ellipse/ && do {
-		$pm->draw_arc($rgc, FALSE, @rb, 0, 64*360) if @rb;
-		my $a = abs(floor($w * sqrt(2)));
-		my $b = abs(floor($h * sqrt(2)));
-		@rb = ($begin[0] - $a, $begin[1] - $b, 2*$a, 2*$b);
-		$pm->draw_arc($rgc, FALSE, @rb, 0, 64*360);
-	    };
-	    /polygon/ && do {
-		my @p = @{$self->{path}};
-		my @points;
-		for my $p (@p) {
-		    push @points, @$p;
-		}
-		push @points, @rb if @rb;
-		$pm->draw_polygon($rgc, 1, @points);
-		@rb = @end;
-		pop @points;
-		pop @points;
-		push @points, @rb;
-		$pm->draw_polygon($rgc, 1, @points);
-	    }
-	}
-
-	@{$self->{rubberband}} = @rb;
-	
-	$self->{image}->set_from_pixbuf(undef);
-	$self->{image}->set_from_pixmap($pm, undef);
-	$handled = 1;
-    }
-    
-    $self->signal_emit('motion-notify');
-    return $handled;
-}
-
-## @method @rubberband_value()
-# @brief Computes a value relevant to current rubberband (length or area) in world coordinates.
-# @return ($dimension, $value) $dimension is either 1 or 2
-sub rubberband_value {
-    my($self) = @_;
-
-    if ($self->{path}) {
-
-	my @p0 = $self->event_pixel2point(@{$self->{path}[0]}) if $self->{path}[0];
-	my @p1 = $self->event_pixel2point(@{$self->{event_coordinates}});
-
-	for ($self->{rubberband_geometry}) {
-	    (/line/ || /path/) && do {
-		my $ogc = new Geo::OGC::LinearRing(points => $self->{path});
-		$ogc->AddPoint(Geo::OGC::Point->new(@{$self->{event_coordinates}}));
-		$ogc->ApplyTransformation( sub { return $self->event_pixel2point(@_); } );
-		return (1, $ogc->Length);
-	    };
-	    /rect/ && do {
-		return (2, abs(($p1[0]-$p0[0])*($p1[1]-$p0[1])));
-	    };
-	    /ellipse/ && do {
-		my $a = ($p1[0]-$p0[0]) * sqrt(2);
-		my $b = ($p1[1]-$p0[1]) * sqrt(2);
-		return (2, abs(3.14159266*$a*$b));
-	    };
-	    /polygon/ && do {
-		my $ogc = new Geo::OGC::LinearRing(points => $self->{path});
-		$ogc->AddPoint(Geo::OGC::Point->new(@{$self->{event_coordinates}}));
-		$ogc->ApplyTransformation( sub { return $self->event_pixel2point(@_); } );
-		$ogc->Close;
-		return (2, undef) unless $ogc->IsSimple;
-		return (2, abs($ogc->Area));
-	    };
-	}
-    }
-}
-
-## @method @event_pixel2point(@pixel)
-# @brief Conversion from pixmap (event) coordinates to the world coordinates.
+## @method @pixel2point(@pixel)
+# @brief Conversion from pixmap (event) coordinates to world
+# coordinates. Alternative name: event_pixel2point:
 sub event_pixel2point {
     my($self, @pixel) = @_;
     return unless $self->{layers} and @{$self->{layers}};
@@ -998,14 +1157,25 @@ sub event_pixel2point {
     return ($self->{minX} + $self->{pixel_size} * ($self->{offset}[0] + $pixel[0] + 0.5),
 	    $self->{maxY} - $self->{pixel_size} * ($self->{offset}[1] + $pixel[1] + 0.5));
 }
+*pixel2point = *event_pixel2point;
 
-## @method @point2pixmap_pixel(@point)
-# @brief Conversion from world coordinates to the pixmap coordinates.
+## @method @point2pixel(@point)
+# @brief Conversion from world coordinates to pixmap
+# coordinates. Alternative name: point2pixmap_pixel.
 sub point2pixmap_pixel {
     my($self, @p) = @_;
     return (round(($p[0] - $self->{minX})/$self->{pixel_size} - 0.5 - $self->{offset}[0]),
 	    round(($self->{maxY} - $p[1])/$self->{pixel_size} - 0.5 - $self->{offset}[1]));
 }
+*point2pixel = *point2pixmap_pixel;
+
+## @method @point2surface(@point)
+# @brief Conversion from world coordinates to surface coordinates (as used in Cairo).
+sub point2surface {
+    my($self, @p) = @_;
+    return ((($p[0] - $self->{minX})/$self->{pixel_size} - $self->{offset}[0]),
+	    (($self->{maxY} - $p[1])/$self->{pixel_size} - $self->{offset}[1]));
+}
 
 ## @ignore
 sub min {
